(require 'org-latex-preview)
(require 'dash)
(require 'tex)

(defun latex-latex-preview--make-preamble ()
  "Slurp the preamble from the `TeX-master-file'.
Code stolen from `TeX-region-create'."
  (save-excursion
    (save-restriction
      (set-buffer                       ; TeX master
       (find-file-noselect (TeX-master-file TeX-default-extension)))
      (save-excursion
        (save-restriction
          (widen)
          (goto-char (point-min))
          (if (not (re-search-forward TeX-header-end nil t))
              ""
            (re-search-backward TeX-header-end nil t)
            (string-replace
             "\\pdfoutput=1"            ; HACK
             ""
             (buffer-substring-no-properties (point-min) (point)))))))))

(defun latex-latex-preview--matching-end ()
  "Find the end of the current environment.
Like `LaTeX-find-matching-end', but take care of corner cases
like being at the very beginning/end of the current environment."
  (let ((boi (save-excursion (LaTeX-back-to-indentation) (point))))
    (cond ((equal (point) boi) (forward-char))
          ((pos-eol)           (backward-char)))
    (LaTeX-find-matching-end)))

(defun latex-latex-preview--math-bounds ()
  "Bet bounds for maths environment at point, if any.
Return a list with (BEG END)."
  (save-excursion
    ;; `texmathp' wants us to be *inside* a maths env.
    (save-match-data
      (when (looking-at org-latex-preview--tentative-math-re)
        (goto-char (match-end 0))))
    (-if-let* (((mm . beg) (and (texmathp) texmathp-why)))
        (list beg
              (and (pcase mm
                     ;; Horrible special casing that should be removed. Also, it
                     ;; *is* possible to mix \( with $ etc., but I'm not sure I
                     ;; even want to allow that.
                     ("$" (search-forward "$" nil t))
                     ("$$" (search-forward "$$" nil t))
                     ("\\[" (search-forward "\\]" nil t))
                     ("\\(" (search-forward "\\)" nil t))
                     (_ (latex-latex-preview--matching-end)))
                   (point))))))

(defun latex-latex-preview-collect-fragments (&optional beg end)
  "Collect LaTeX maths fragments between BEG and END.
Lots stolen from `org-latex-preview-collect-fragments'."
  ;; TODO: This does not collect environments at this point, only maths
  ;; fragments.
  (let (fragments)
    (save-excursion
      (goto-char (or beg (point-min)))
      (while (re-search-forward org-latex-preview--tentative-math-re end t)
        (when-let ((bounds (latex-latex-preview--math-bounds))
                   ;; Avoid duplicating nested latex environments.
                   (_ (not (and fragments (= (car bounds) (caar fragments))))))
          (push bounds fragments))))
    (nreverse fragments)))

(defun latex-latex-preview-place (entries &optional preamble)
  "Preview LaTeX maths fragments ENTRIES.
See `org-latex-preview-place'."
  (org-latex-preview-place
   org-latex-preview-process-default
   entries
   nil
   (or preamble (latex-latex-preview--make-preamble))))

;;;###autoload
(defun latex-latex-preview-region (beg end)
  "Preview all LaTeX fragments in the active region."
  (interactive "r")
  (latex-latex-preview-place
   (latex-latex-preview-collect-fragments beg end)
   (latex-latex-preview--make-preamble)))

;;;###autoload
(defun latex-latex-preview ()
  "Preview LaTeX fragment at point."
  (interactive)
  (latex-latex-preview-place
   (list (latex-latex-preview--math-bounds))
   (latex-latex-preview--make-preamble)))

(defun latex-latex-preview-clear-cache (beg end)
  "A primitive version of `org-latex-preview-clear-cache'.
TODO: clear-entire-cache."
  (interactive
   (cond ((use-region-p)
          (list (region-beginning) (region-end)))
         ((latex-latex-preview--math-bounds))
         (t (error "latex-latex-preview-clear-cache needs something to clear"))))
  (org-latex-preview-clear-overlays beg end)
  (let ((imagetype
         (or (plist-get (alist-get org-latex-preview-process-default
                                   org-latex-preview-process-alist)
                        :image-output-type)
             "png"))
        ;; TODO: numbered previews
        )
    ;; TODO: This does not quite work yet.
    (pcase-dolist (`(,begin ,end) (latex-latex-preview-collect-fragments beg end))
      (pcase-let* ((`(,fg ,bg) (org-latex-preview--colors-around begin end))
                   (value (buffer-substring begin end)))
        (org-latex-preview--remove-cached
         (org-latex-preview--hash
          org-latex-preview-process-default
          org-latex-preview--preamble-content
          value imagetype fg bg nil))))
    (message "Cleared LaTeX preview cache for %s."
             (if (or beg end) "region" "buffer"))))

;;; Auto previewing.

(defun latex-latex-preview-auto--regenerate-overlay (ov &optional inhibit-renumbering)
  "Regenerate the LaTeX fragment under overlay OV.
See `org-latex-preview-auto--regenerate-overlay'.
TODO: equation numbering"
  (with-current-buffer (overlay-buffer ov)
    (if-let* ((start (overlay-start ov))
              (end (overlay-end ov))
              (fragment (save-excursion
                          (goto-char start)
                          (latex-latex-preview--math-bounds))))
        (if (org-latex-preview--empty-fragment-p ; no org-specific code
             (buffer-substring-no-properties start end))
            (progn (delete-overlay ov)
                   (org-latex-preview--ensure-overlay start end)) ; no org-specific code
          (latex-latex-preview-place (list fragment)))
      ;; Delete overlay of fragment is no longer in maths mode.
      (delete-overlay ov))))

;;;###autoload
(define-minor-mode latex-latex-preview-auto-mode
  "See `org-latex-preview-auto-mode'."
  :global nil
  (if latex-latex-preview-auto-mode
      (progn
        (setq org-latex-preview-auto--marker (make-marker))
        ;; This allows us to reuse the code from
        ;; `org-latex-preview-auto--close-previous-overlay' without having to
        ;; copy the whole function.
        (advice-add 'org-latex-preview-auto--regenerate-overlay :override
          #'latex-latex-preview-auto--regenerate-overlay)
        (add-hook 'pre-command-hook #'org-latex-preview-auto--handle-pre-cursor nil 'local)
        (org-latex-preview-auto--handle-pre-cursor) ; Invoke setup before the hook even fires.
        (add-hook 'post-command-hook #'org-latex-preview-auto--handle-post-cursor nil 'local)
        ;; TODO: Live preview
        ;; (add-hook 'after-change-functions #'org-latex-preview-auto--detect-fragments-in-change nil 'local)
        ;; (when org-latex-preview-live
        ;;   (org-latex-preview-live--setup))
        )
    (advice-remove 'org-latex-preview-auto--regenerate-overlay
                   'latex-latex-preview-auto--regenerate-overlay)
    (remove-hook 'pre-command-hook #'org-latex-preview-auto--handle-pre-cursor 'local)
    (remove-hook 'post-command-hook #'org-latex-preview-auto--handle-post-cursor 'local)
    ;; (remove-hook 'after-change-functions #'org-latex-preview-auto--detect-fragments-in-change 'local)
    ;; (org-latex-preview-live--teardown)
    ))

(provide 'latex-latex-preview)
;;; latex-latex-preview.el ends here
